/*******************************************************************************
 * Copyright (c) 2007, 2019 Association for Decentralized Information Management
 * in Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.db.layer0.adapter.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.common.primitiverequest.Adapter;
import org.simantics.db.common.procedure.adapter.TransientCacheListener;
import org.simantics.db.common.request.ObjectsWithType;
import org.simantics.db.common.request.TernaryRead;
import org.simantics.db.exception.DatabaseException;
import org.simantics.db.layer0.adapter.Instances;
import org.simantics.db.layer0.genericrelation.Dependencies;
import org.simantics.db.layer0.genericrelation.IndexQueries;
import org.simantics.db.layer0.util.Layer0Utils;
import org.simantics.db.request.ReadExt;
import org.simantics.db.request.RequestFlags;
import org.simantics.db.service.CollectionSupport;
import org.simantics.layer0.Layer0;
import org.simantics.operation.Layer0X;
import org.simantics.scl.runtime.function.Function;

import gnu.trove.set.hash.THashSet;

/**
 * @author Antti Villberg
 * @author Tuukka Lehtonen
 */
public class EntityInstances implements Instances {

    private static final boolean TRACE_QUERIES = false;

    private final Resource type;

    public EntityInstances(Resource type) {
        this.type = type;
    }

    @Override
    public Collection<Resource> find(ReadGraph graph, Resource index) throws DatabaseException {
        return find(graph, index, "");
    }

    /**
     * A (cacheable) query to optimize single index queries for immutable
     * indexes such as ontologies.
     */
    public static class QueryIndex extends TernaryRead<Resource, Resource, String, List<Resource>> implements ReadExt {

        public QueryIndex(Resource index, Resource type, String filter) {
            super(index, type, filter);
        }

        @Override
        public List<Resource> perform(ReadGraph graph)
                throws DatabaseException {
            Resource type = parameter2;
            String filter = constructLuceneQuery(graph, type, parameter3);
            if (filter == null)
                return Collections.emptyList();

            @SuppressWarnings({ "unchecked", "rawtypes" })
            Function dependencyResources = graph.syncRequest(new Adapter(Layer0X.getInstance(graph).DependencyResources, Function.class), TransientCacheListener.<Function>instance());

            if (TRACE_QUERIES) {
                System.out.println("EntityInstances.QueryIndex: finding " + filter + " from index " + graph.getPossibleURI(parameter));
                //new Exception("EntityInstances: finding " + filter + " from index " + graph.getPossibleURI(parameter)).printStackTrace();
            }

            @SuppressWarnings("unchecked")
            List<Resource> results = (List<Resource>)dependencyResources.apply(graph, parameter, filter);
            if (results == null || results.isEmpty())
                return Collections.emptyList();

            if (TRACE_QUERIES)
                System.out.println("  EntityInstances.QueryIndex: got " + results.size() + " results");

            // Optimize single result case.
            if (results.size() == 1) {
                Resource singleResult = results.get(0);
                List<Resource> result = graph.isInstanceOf(singleResult, type) ? results : Collections.emptyList();
                if (TRACE_QUERIES)
                    System.out.println("  EntityInstances.QueryIndex: got " + results.size() + " unique type-matching results");
                return result;
            }

            CollectionSupport coll = graph.getService(CollectionSupport.class);
            List<Resource> result = coll.createList(results.size());
            for (Resource res : Layer0Utils.sortByCluster(graph, results))
                if (graph.isInstanceOf(res, type))
                    result.add(res);

            if (TRACE_QUERIES)
                System.out.println("  EntityInstances.QueryIndex: got " + results.size() + " unique type-matching results");

            return result;
        }

        private String constructLuceneQuery(ReadGraph graph, Resource type, String filter) throws DatabaseException {
            Layer0 L0 = Layer0.getInstance(graph);
            StringBuilder sb = new StringBuilder();
            boolean emptyFilter = filter.isEmpty();
            if (emptyFilter || !type.equals(L0.Entity)) {
                IndexQueries.appendResourceIdTerm(sb, Dependencies.FIELD_TYPE_RESOURCE, type);
            }
            if (!emptyFilter) {
                if (sb.length() > 0)
                    sb.append(" AND ");
                sb.append(filter);
            }
            if (sb.length() == 0) {
            	sb.append("*:*");
            }
            return sb.toString();
        }

        @Override
        public String toString() {
        	return "QueryIndex " + parameter + " " + parameter2 + " " + parameter3;
        }

        @Override
        public boolean isImmutable(ReadGraph graph) throws DatabaseException {
            return graph.isImmutable(parameter);
        }

		@Override
		public int getType() {
			// This query should not be immediate update since it takes a long time!
			return RequestFlags.INVALIDATE;
		}

    }

    private List<Resource> findRec(ReadGraph graph, Resource index, String filter, Set<Resource> visited) throws DatabaseException {

    	if(!visited.add(index)) return Collections.emptyList();

    	CollectionSupport coll = graph.getService(CollectionSupport.class);

    	List<Resource> indexResult = graph.syncRequest(new QueryIndex(index, type, filter), TransientCacheListener.<List<Resource>>instance());

        Layer0 L0 = Layer0.getInstance(graph);
        Collection<Resource> linkedRoots = graph.syncRequest(new ObjectsWithType(index, L0.IsLinkedTo, L0.IndexRoot));
        if (linkedRoots.isEmpty())
            return indexResult;

        List<Resource> result = indexResult;
        for (Resource dep : linkedRoots) {
            Collection<Resource> linkedIndexResults = findRec(graph, dep, filter, visited);
            if (linkedIndexResults.isEmpty())
                continue;
            if (result == indexResult) {
            	result = coll.createList();
                result.addAll(indexResult);
            } else {
            }
            result.addAll(linkedIndexResults);
        }
        
        return result;
        
    }

    @Override
    public Collection<Resource> find(ReadGraph graph, Resource index, String filter) throws DatabaseException {
    	CollectionSupport coll = graph.getService(CollectionSupport.class);
    	
    	THashSet<Resource> visited = new THashSet<>();
    	List<Resource> rec_ = findRec(graph, index, filter, visited);
    	// We must not modify rec_!
    	List<Resource> rec = rec_;
    	for(Resource global : Layer0Utils.listGlobalOntologies(graph)) {
        	if(!visited.add(global)) continue;
    		List<Resource> rs = graph.syncRequest(new QueryIndex(global, type, filter), TransientCacheListener.<List<Resource>>instance());
    		if(!rs.isEmpty()) {
    		    if(rec == rec_)
    		        rec = new ArrayList<>(rec);
                rec.addAll(rs);
    		}
    	}
    	Collection<Resource> result = coll.asSortedList(rec);
    	return result; 
    }
    
    @Override
    public Collection<Resource> findByName(ReadGraph graph, Resource model,
            String name) throws DatabaseException {
        return find(graph, model, IndexQueries.quoteTerm(name));
    }

}
